from typing import Dict, Any, Optional, Tuple
import yaml

from .models import LoraConfig, TrainerConfig


class PromptHandler:
    
    @staticmethod
    def set_prompt(doc: Dict[str, Any]) -> str:
        """Constructs a cypher statement based on the given document.

        Args:
            doc (Dict[str, Any]): A dictionary containing schema, input, and output keys.

        Returns:
            str: A formatted cypher statement prompt.
        """
        prompt = f"""Given a sentence in natural language, construct a cypher statement in order to extract information from a knowledge graph.
            The graph will have the following schema:
            {doc["schema"]}

            ### sentence:
            {doc["input"]}
            
            ### cypher statement:
            {doc["output"]}
            """
        return prompt
    
class YamlFileManager:
    @staticmethod
    def parse_yaml_file(yaml_file_path: str) -> Optional[Tuple[LoraConfig, TrainerConfig]]:
        """Parse a YAML file and return a tuple of LoraConfig and TrainerConfig objects.

        Args:
            yaml_file_path (str): The path to the YAML file.

        Returns:
            Optional[Tuple[LoraConfig, TrainerConfig]]: A tuple containing LoraConfig and TrainerConfig objects, or None on error.
        """
        try:
            with open(yaml_file_path, "r") as yaml_file:
                yaml_content = yaml.safe_load(yaml_file)

            lora_config = LoraConfig(**yaml_content["LoraConfig"])
            trainer_config = TrainerConfig(**yaml_content["TrainerConfig"])

            return lora_config, trainer_config

        except (FileNotFoundError, yaml.YAMLError) as e:
            print(f"Error parsing YAML file: {e}")
            return None



import os
import torch
from datasets import load_dataset
from peft import prepare_model_for_kbit_training, LoraConfig, get_peft_model
import wandb
import transformers
from transformers import (
    AutoTokenizer, 
    AutoModelForCausalLM, 
    BitsAndBytesConfig, 
    TrainingArguments, 
    DataCollatorForLanguageModeling
)
from .utils import PromptHandler, YamlFileManager as manager
from accelerate import FullyShardedDataParallelPlugin, Accelerator
from torch.distributed.fsdp.fully_sharded_data_parallel import FullOptimStateDictConfig, FullStateDictConfig

fsdp_plugin = FullyShardedDataParallelPlugin(
    state_dict_config=FullStateDictConfig(offload_to_cpu=True, rank0_only=False),
    optim_state_dict_config=FullOptimStateDictConfig(offload_to_cpu=True, rank0_only=False),
)

class CypherTuner:
    """
    A class for fine-tuning an LLM using QLoRA.

    Args:
        project_name: Name of the project for organization and tracking.
        model_id: Identifier for the base model from Hugging Face.
        dataset_id: Identifier for the dataset used for training and evaluation.
        config_file: path to YAML config file.
    """
    def __init__(self, project_name, model_id, dataset_id, config_file):

        self.project_name = project_name
        self.model_id = model_id
        self.dataset_id = dataset_id

        # Parse configuration file
        lora_config, trainer_config = manager.parse_yaml_file(config_file)

        # Use the parsed configurations
        self.lora_config = LoraConfig(**lora_config.model_dump())
        self.args = TrainingArguments(**trainer_config.model_dump())
        
        self.max_length = 340
        
        self.bnb_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_use_double_quant=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_compute_dtype=torch.bfloat16,
        )
        
        self.accelerator = Accelerator(fsdp_plugin=fsdp_plugin)

    def load_datasets(self):
        """
        Load the training and evaluation datasets from the specified dataset ID.
        """
        train_dataset = load_dataset(path=self.dataset_id, split="train")
        eval_dataset = load_dataset(path=self.dataset_id, split="validation")
        
        return train_dataset, eval_dataset

    def load_model_and_tokenizer(self):
        """
        Load the base model and tokenizer for the Llama 2 7B model from Hugging Face.
        """
        model = AutoModelForCausalLM.from_pretrained(
            self.model_id, quantization_config=self.bnb_config, trust_remote_code=True,
        )

        tokenizer = AutoTokenizer.from_pretrained(
            self.model_id,
            padding_side="left",
            add_eos_token=True,  
            add_bos_token=True,  
        )
        
        return model, tokenizer
    
    def create_prompts_from_datasets(self, tokenizer, train_data, eval_data):
        """
        Create and tokenize prompts from the datasets.

        Args:
            tokenizer: The tokenizer to be used for tokenizing the prompts.
            train_data: The training dataset.
            eval_data: The evaluation dataset.

        Returns:
            Tuple of tokenized training and evaluation datasets.
        """
        tokenizer.pad_token = tokenizer.eos_token

        def tokenize(prompt):
            result = tokenizer(
                prompt,
                truncation=True,
                max_length=self.max_length,
                padding="max_length",
            )
            result["labels"] = result["input_ids"].copy()
            
            return result
        
        def generate_and_tokenize_prompt(doc):
            return tokenize(PromptHandler.set_prompt(doc))
        
        tokenized_train_dataset = train_data.map(generate_and_tokenize_prompt)
        tokenized_eval_dataset = eval_data.map(generate_and_tokenize_prompt)

        return tokenized_train_dataset, tokenized_eval_dataset

    def configure_lora(self, model):
        """
        Configure the model with Lora settings for fine-tuning.

        Args:
            model: The model to be configured.

        Returns:
            The configured model.
        """
        model.gradient_checkpointing_enable()
        model = prepare_model_for_kbit_training(model)
        model = get_peft_model(model, self.lora_config)
        model = self.accelerator.prepare_model(model)
        return model

    def configure_training(self, model, tokenizer, train_dataset, eval_dataset):
        """
        Set up the training configuration using Transformers Trainer.

        Args:
            model: The model to be trained.
            tokenizer: The tokenizer to be used during training.
            train_dataset: The dataset to be used for training.
            eval_dataset: The dataset to be used for evaluation.

        Returns:
            Configured trainer object.
        """
        if torch.cuda.device_count() > 1:
            model.is_parallelizable = True
            model.model_parallel = True
        tokenizer.pad_token = tokenizer.eos_token

        trainer = transformers.Trainer(
            model=model,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            args=self.args,
            data_collator=DataCollatorForLanguageModeling(tokenizer, mlm=False),
        )

        model.config.use_cache = False
        wandb.login()
        wandb_project = "tllm-finetune"
        if len(wandb_project) > 0:
            os.environ["WANDB_PROJECT"] = wandb_project

        return trainer
    
    def train(self):
        """
        Prepares datasets, model, tokenizer, and configurations, and then starts the training process.
        """
        
        print("Preparing your training job...")

        train_data, eval_data = self.load_datasets()
        model, tokenizer = self.load_model_and_tokenizer()
        train_data, eval_data = self.create_prompts_from_datasets(tokenizer, train_data, eval_data)
        model = self.configure_lora(model)
        trainer = self.configure_training(model, tokenizer, train_data, eval_data)
        trainer.train()


import torch
from peft import PeftModel


class InferenceHandler:
    def load_finetuned_model(self, checkpoint_path):
        """
        Load the fine-tuned model checkpoint.

        Args:
            checkpoint_path (str): The path to the checkpoint directory.
        """
        self.ft_model = PeftModel.from_pretrained(self.base_model, checkpoint_path)

    def generate_text(self, eval_prompt):
        """
        Generate text using the fine-tuned model.

        Args:
            eval_prompt (str): The prompt for text generation.

        Returns: 
            str: The generated text.
        """
        model_input = self.tokenizer(eval_prompt, return_tensors="pt").to("cuda")
        self.ft_model.eval()
        with torch.no_grad():
            return self.tokenizer.decode(
                self.ft_model.generate(**model_input, max_new_tokens=300)[0],
                skip_special_tokens=True,
            )


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    LONG_DESCRIPTION = "\n" + fh.read()

NAME = "cyphertune"
VERSION = "0.0.0.3"
AUTHOR = "InquestGeronimo"
EMAIL = "rcostanl@gmail.com"
LD_CONTENT_TYPE = "text/markdown"
DESCRIPTION = "A Trainer for Fine-tuning LLMs for Text-to-Cypher Datasets"
LICENSE = "Apache 2.0 license"
PACKAGES = find_packages()
DEPENDENCIES = [
    "accelerate>=0.27.0",
    "bitsandbytes>=0.42.0",
    "datasets>=2.17.0",
    "peft>=0.8.2",
    "transformers>=4.37.2",
    "wandb>=0.16.3",
    "scipy>=1.11.4",
    "pydantic>=2.6.1"
]
KEYWORDS = ["llms", "training", "fine-tuning", "LLM", "NLP"]
CLASSIFIERS = [
    "Development Status :: 1 - Planning",
    "Intended Audience :: Developers",
    "Programming Language :: Python :: 3",
    "Operating System :: Unix",
]

setup(
    name=NAME,
    version=VERSION,
    license=LICENSE,
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description_content_type=LD_CONTENT_TYPE,
    long_description=LONG_DESCRIPTION,
    packages=PACKAGES,
    include_package_data=True,
    install_requires=DEPENDENCIES,
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
)



from pydantic import BaseModel
from typing import List

class LoraConfig(BaseModel):
    r: int
    lora_alpha: int
    target_modules: List[str]
    bias: str
    lora_dropout: float
    task_type: str


class TrainerConfig(BaseModel):
    output_dir: str
    warmup_steps: int
    per_device_train_batch_size: int
    gradient_accumulation_steps: int
    gradient_checkpointing: bool
    max_steps: int
    learning_rate: float
    bf16: bool
    optim: str
    logging_dir: str
    save_strategy: str
    save_steps: int
    evaluation_strategy: str
    eval_steps: int
    do_eval: bool
    report_to: str
    remove_unused_columns: bool
    run_name: str